/* -*- C -*-
 * $Id$
 *
 * @author Daniel Silverstone <dsilvers@digital-scurf.org>
 */

/*
 * Implement a LRU cache of MAXOBJECTS entries.
 * Any entry marked as *IMPORTANT* goes into a different cache
 */

#include <pthread.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include <fcntl.h>
#include "settings.h"
#include <stdio.h>

#include <lua.h>

#define MAXOBJECTS   64
#define MAXIMPORTANT 16
#define MAXHITS      10
#define MAXAGE       60

unsigned int hits = 0;
unsigned int miss = 0;

int skip = 0;

unsigned int caffeine(const char* str)
{
  unsigned int ret = 0x0c0ffee0;
  str += skip;
  DebugOut("The length of str is %d, having skipping %d and ",strlen(str), skip);
  DebugOut("the Caffeine of \"%s\" is",str);
  while (*str)
    {
      ret = ret ^ *str++;
      ret = (ret << 17) | (ret >> (32-17));
      ret = ret ^ 0xdeadbeef;
    }
  DebugOut(" %x.\n",ret);
  return ret;
}

typedef struct 
{
  unsigned int hash;
  char *name;
  unsigned char *data;
  unsigned char *comp;
  int size;
  int csiz;
  int mtime;
  int atime;
  int hits;
} CacheEntry;

static CacheEntry ObjectCache[MAXOBJECTS];
static CacheEntry ImportantCache[MAXIMPORTANT];

#define WhichCache(i) (i?(&ImportantCache[0]):(&ObjectCache[0]))
#define MaxEntries(i) (i?MAXIMPORTANT:MAXOBJECTS)

void InitCaches(void)
{
  int i;
  for(i=0; i < MAXOBJECTS; ++i)
    {
      ObjectCache[i].hash = 0;
      ObjectCache[i].name = 0;
      ObjectCache[i].data = 0;
      ObjectCache[i].comp = 0;
      ObjectCache[i].size = 0;
      ObjectCache[i].csiz = 0;
      ObjectCache[i].mtime = 0;
      ObjectCache[i].atime = 0;
      ObjectCache[i].hits = 0;
    }
  for(i=0; i < MAXIMPORTANT; ++i)
    {
      ImportantCache[i].hash = 0;
      ImportantCache[i].name = 0;
      ImportantCache[i].data = 0;
      ImportantCache[i].comp = 0;
      ImportantCache[i].size = 0;
      ImportantCache[i].csiz = 0;
      ImportantCache[i].mtime = 0;
      ImportantCache[i].atime = 0;
      ImportantCache[i].hits = 0;
    }
}

void ClearCaches(void)
{
  int i;
  for(i=0; i < MAXOBJECTS; ++i)
    {
      ObjectCache[i].hash = 0;
      if (ObjectCache[i].name)
	free(ObjectCache[i].name);
      ObjectCache[i].name = 0;
      if (ObjectCache[i].data)
	free(ObjectCache[i].data);
      ObjectCache[i].data = 0;
      if (ObjectCache[i].comp)
	free(ObjectCache[i].comp);
      ObjectCache[i].comp = 0;
      ObjectCache[i].size = 0;
      ObjectCache[i].csiz = 0;
      ObjectCache[i].atime = 0;
      ObjectCache[i].hits = 0;
    }
  for(i=0; i < MAXIMPORTANT; ++i)
    {
      ImportantCache[i].hash = 0;
      if (ImportantCache[i].name)
	free(ImportantCache[i].name);
      ImportantCache[i].name = 0;
      if (ImportantCache[i].data)
	free(ImportantCache[i].data);
      ImportantCache[i].data = 0;
      if (ImportantCache[i].comp)
	free(ImportantCache[i].comp);
      ImportantCache[i].comp = 0;
      ImportantCache[i].size = 0;
      ImportantCache[i].csiz = 0;
      ImportantCache[i].mtime = 0;
      ImportantCache[i].atime = 0;
      ImportantCache[i].hits = 0;
    }
}

int FindInCache(const char* filename, int important)
{
  CacheEntry *c = WhichCache(important);
  unsigned int target = caffeine(filename);
  int i;
  for(i=0;i<MaxEntries(important);++i)
    {
      if (c[i].hash == target)
	if (!strcmp(c[i].name,filename))
	  return i;
    }
  return -1;
}

int FindSpace(int important)
{
  CacheEntry *c = WhichCache(important);
  int oldest = -1;
  int oldtime = INT_MAX;
  int i;
  for(i=0;i<MaxEntries(important);++i)
    {
      if (c[i].hash == 0)
	return i;
      if (c[i].atime < oldtime)
	{
	  oldtime = c[i].atime;
	  oldest = i;
	}
    }
  return oldest;
}

const unsigned char * GetData(int ofs,int important)
{
  return WhichCache(important)[ofs].data;
}

const unsigned char * GetComp(int ofs,int important)
{
  return WhichCache(important)[ofs].comp;
}

const int GetSize(int ofs, int important)
{
  return WhichCache(important)[ofs].size;
}

const int GetCompSize(int ofs, int important)
{
  return WhichCache(important)[ofs].csiz;
}

const int GetMTime(int ofs, int important)
{
  return WhichCache(important)[ofs].mtime;
}

void Insert(const char* filename, CacheEntry *c, int p)
{
  struct stat sbuf;
  int fd;
  if (c[p].hash)
    {
      /* It was already used, clean it up */
      free(c[p].data);
      free(c[p].name);
      if (c[p].comp)
	free(c[p].comp);
    }
  
  c[p].hash = caffeine(filename);
  c[p].name = (char *)(malloc(strlen(filename)+1));
  strcpy(c[p].name,filename);
  if (stat(filename,&sbuf))
    {
      c[p].data = 0;
      c[p].comp = 0;
      c[p].size = -1;
      c[p].csiz = -1;
      c[p].mtime = 0;
    }
  else
    {
      c[p].size = sbuf.st_size;
      c[p].mtime = sbuf.st_mtime;
      c[p].data = (unsigned char*)(malloc(sbuf.st_size));
      fd = open(filename,O_RDONLY);
      read(fd,c[p].data,c[p].size);
      c[p].comp = 0;
      c[p].csiz = -1;
      close(fd);
    }
  c[p].atime = time(0);
  c[p].hits = 1;
}

void CheckAge(CacheEntry *c,int p)
{
  struct stat sbuf;
  int fd;
  if (stat(c[p].name,&sbuf))
    {
      /* File lost underneath us, so we blank it out */
      DebugOut("Unfortunately we lost it\n");
      c[p].atime = time(0);
      c[p].mtime = 0;
      if (c[p].data)
	free(c[p].data);
      c[p].data = 0;
      if (c[p].comp)
	free(c[p].comp);
      c[p].comp = 0;
      c[p].size = -1;
      c[p].csiz = -1;
      c[p].hits = 0;
      return;
    }
  /* Didn't lose file, so do we need to reload it? */
  if (((time(0) - c[p].atime) > MAXAGE) || (c[p].hits > MAXHITS));
    {
      DebugOut("Worth thinking about...\n");
      if (sbuf.st_mtime != c[p].mtime)
	{
	  DebugOut("Coo, it was changed\n");
	  /* It has been modified */
	  if (c[p].data)
	    free(c[p].data);
	  if (c[p].comp)
	    free(c[p].comp);
	  c[p].comp = 0;
	  c[p].csiz = -1;
	  c[p].size = sbuf.st_size;
	  c[p].mtime = sbuf.st_mtime;
	  c[p].data = (unsigned char*)(malloc(sbuf.st_size));
	  c[p].atime = time(0);
	  fd = open(c[p].name,O_RDONLY);
	  read(fd,c[p].data,c[p].size);
	  close(fd);
	  c[p].hits = 0;
	  hits--;miss++; /* Adjust stats, given we missed to disk */
	}
    }

}

int GetFile(const char* filename, int important)
{
  int p = FindInCache(filename,important);
  CacheEntry *c = WhichCache(important);
  if (p > -1)
    {
      hits++;
      DebugOut("Hit '%s' in cache...\n",filename);
      CheckAge(c,p); /* The entry is up-to-date */
      c[p].hits++;
      return p;
    }
  miss++;
  p = FindSpace(important);
  DebugOut("Added '%s' to the%s cache\n",filename,important?" IMPORTANT FILES":"");
  Insert(filename,c,p);
  return p;
}

pthread_mutex_t cachemutex = PTHREAD_MUTEX_INITIALIZER;
void LockCache()
{
  DebugOut("Lock Cache\n");
  pthread_mutex_lock(&cachemutex);
}

void UnlockCache()
{
  pthread_mutex_unlock(&cachemutex);
  DebugOut("Unlock Cache\n");
}

int LoadCacheFile(lua_State *L)
{
  const char* name;
  int important;
  const char *data;
  int length;
  int cacheentry;
  name = lua_tostring(L,1);
  important = lua_tonumber(L,2);
  DebugOut("Being asked to load '%s' into the %s\n",name,important?"IMPORTANT cache":"cache");
  if (!name)
    {
      lua_error(L,"Cannot load NULL");
    }
  LockCache();
  cacheentry = GetFile(name,important);
  length = GetSize(cacheentry,important);
  data = GetData(cacheentry,important);
  if (length == -1)
    lua_pushnil(L);
  else
    lua_pushlstring(L,data,length);
  UnlockCache();
  return 1; /* the string */
}

pthread_mutex_t compilemutex = PTHREAD_MUTEX_INITIALIZER;

void AttemptCompile(int ce, int imp)
{
  /* Flow of operations:
   *
   * 1. Get source ptr
   * 2. get ZIO ptr for the memory ptr
   * 3. get a lua_State suitable for compiling in.
   * 4. enter a "protected" section
   * 5. parse the file into a Proto
   * 6. dump the proto to a memory block
   * 7. Free the proto and associated memory somehow (I think this happens on lua_close)
   * 8. remove the lua_State
   * 9. Update cache
   */
  pthread_mutex_lock(&compilemutex);


  pthread_mutex_unlock(&compilemutex);
}

int DoCacheFile(lua_State *L)
{
  const char* name;
  int important;
  const char *data;
  const char *comp;
  int compsize;
  int length;
  int cacheentry;
  const char *mycopy;
  name = lua_tostring(L,1);
  important = lua_tonumber(L,2);
  DebugOut("Being asked to run '%s' from the %s\n",name,important?"IMPORTANT cache":"cache");
  if (!name)
    {
      lua_error(L,"Cannot run NULL");
    }
  LockCache();
  /*
  cacheentry = GetFile(name,important);
  length = GetSize(cacheentry,important);
  comp = GetComp(cacheentry,important);
  if (length == -1)
    lua_pushnil(L);
  else
    {
      if (!comp)
	{
	  AttemptCompile(L,cacheentry,important); 
	  comp = GetComp(cacheentry,important);
	}
      compsize = GetCompSize(cacheentry,important);
      lua_pushlstring(comp,compsize); 
      mycopy = lua_tostring(L,-1);    
    }
  */
  cacheentry = GetFile(name,important);
  length = GetSize(cacheentry,important);
  if (length > 0)
    {
      lua_pushlstring(L,GetData(cacheentry,important),length);
      mycopy = lua_tostring(L,-1);
    }
  UnlockCache();
  if (length>0)
    lua_dobuffer(L,mycopy,length,name);

  return 0;
}


void register_cache(lua_State *L)
{
  lua_pushcclosure(L,LoadCacheFile,0);
  lua_setglobal(L,"_LoadCache");

  lua_pushcclosure(L,DoCacheFile,0);
  lua_setglobal(L,"_DoCacheFile");
}

