/*
 * $Id$
 */

#include "settings.h"

#include <libpq-fe.h>
#include <lua.h>
#include <stdlib.h>

typedef struct {
  int structtag;
  int connectiontag;
  int resulttag;
} PQTags;

int PGConnect(lua_State *L);

int PGTagCollect(lua_State *L);

void register_postgres(lua_State *L)
{
  PQTags *p = (PQTags*)(malloc(sizeof(PQTags)));
  int t = lua_newtag(L); /* Only for the PQTags structure */
  p->connectiontag = lua_newtag(L);
  p->resulttag = lua_newtag(L);
  p->structtag = t;
  /* GC for the tags structure */
  lua_pushusertag(L, p, t);
  lua_pushcclosure(L, PGTagCollect, 1);
  lua_settagmethod(L, t, "gc");

  /* GC for connections */
  lua_pushusertag(L, p, t);
  lua_pushcclosure(L, PGTagCollect, 1);
  lua_settagmethod(L, p->connectiontag, "gc");

  /* GC for Result objects */
  lua_pushusertag(L, p, t);
  lua_pushcclosure(L, PGTagCollect, 1);
  lua_settagmethod(L, p->resulttag, "gc");
  
  /* Only Globally registered function PGConnect */
  lua_pushusertag(L, p, t);
  lua_pushcclosure(L, PGConnect, 1);
  lua_setglobal(L, "PGConnect");
}

int PGTagCollect(lua_State *L)
{
  /* Called when we have reached the end of the line */
  int tag = lua_tag(L,1);
  PQTags *t = (PQTags *)(lua_touserdata(L,-1));
  DebugOut("Garbage collecting Postgresql Tag information.\n");
  if (tag == t->structtag)
    {
      DebugOut("Struct cleared\n");
      free(lua_touserdata(L,1));
    }
  else if (tag == t->connectiontag)
    {
      DebugOut("Connection cleared\n");
      PQfinish((PGconn*)lua_touserdata(L,1));
    }
  else if (tag == t->resulttag)
    {
      DebugOut("Result cleared\n");
      PQclear((PGresult*)lua_touserdata(L,1));
    }
  else
    {
      DebugOut("Garbage collected unknown tag (%d).\n",tag);
    }
  return 0; /* No results */
}

int PGDisconnect(lua_State *L)
{
  /* Called to disconnect a lua postgres connection */
  /* Param 1 is the table, containing the PGConn */
  /* Param -1 is the PQTags structure defining the tags */
  PQTags *t = (PQTags*)(lua_touserdata(L,-1));
  lua_pushstring(L,"PGconn");
  lua_gettable(L,1);
  /* Top of stack is now the PGConn userdata object */
  if (lua_tag(L,-1) != t->connectiontag)
    {
      lua_error(L,"Unable to close a non-Postgres connection");
    }
  /* Get the Connection and close it in one fell swoop */
  /*PQfinish((PGconn*)(lua_touserdata(L,-1)));*/
  /* and now, remove that from the stack, cleaning up */
  lua_pop(L,1);
  lua_pushstring(L,"PGconn");
  lua_pushnil(L);
  lua_settable(L,1);
  /* and tell Lua we return nothing */
  return 0;
}

int PGExecute(lua_State *L);

#define GRABPARAM(param) lua_pushstring(L,#param);lua_gettable(L,1);param=lua_tostring(L,-1);lua_pop(L,1)


int PGConnect(lua_State *L)
{
  /* Connect to postgresql... */
  /* Parameters: table of params, Tags structure */
  PQTags *t = (PQTags*)(lua_touserdata(L,-1));
  PGconn *c;
  const char *host,*port,*user,*pass,*db,*options,*tty;
  if (lua_gettop(L) != 2)
    {
      lua_error(L,"Incorrect number of parameters to PGConnect");
    }
  if (!lua_istable(L,1))
    {
      lua_error(L,"PGConnect expects a parameter table.");
    }
  GRABPARAM(host);
  GRABPARAM(port);
  GRABPARAM(user);
  GRABPARAM(pass);
  GRABPARAM(db);
  GRABPARAM(options);
  GRABPARAM(tty);
  /* Form connection to database */
  c = PQsetdbLogin(host,port,options,tty,db,user,pass);
  if (!c)
    {
      /* Out of memory condition */
      lua_pushnil(L);
      lua_pushstring(L,"Out of memory - however that happened");
      return 2; /* two things on the stack */
    }
  /* Check the connection formed okay */
  if (PQstatus(c) == CONNECTION_BAD)
    {
      lua_pushnil(L);
      lua_pushstring(L,PQerrorMessage(c));
      PQfinish(c);
      return 2;
    }
  /* Connection okay, build a table... */
  /* Empty the stack */
  lua_settop(L,0);
  lua_newtable(L); /* The return connection table */
  /* First off, get the connection in there */
  lua_pushstring(L,"PGconn");
  lua_pushusertag(L,c,t->connectiontag);
  lua_settable(L,1);
  /* Execute function */
  lua_pushstring(L,"Execute");
  lua_pushusertag(L,t,t->structtag);
  lua_pushcclosure(L,PGExecute,1);
  lua_settable(L,1);
  /* Close Function */
  lua_pushstring(L,"Disconnect");
  lua_pushusertag(L,t,t->structtag);
  lua_pushcclosure(L,PGDisconnect,1);
  lua_settable(L,1);
  /* Other useful data... */

  /* finally, the table is all that is left on the stack, so.. */
  lua_pushnil(L);
  return 2; /* table,nil */
}

int PGDone(lua_State *L)
{
  PQTags *t = (PQTags*)(lua_touserdata(L,-1));

  lua_pushstring(L,"PGresult");
  lua_gettable(L,1);
  
  /*  if (t->resulttag == lua_tag(L,-1))
    {
      PQclear((PGresult*)lua_touserdata(L,-1));
    }
  */
  lua_pop(L,1);
  lua_pushstring(L,"PGresult");
  lua_pushnil(L);
  lua_settable(L,1);
  /* return nothing */
  return 0;
}

int PGGetRow(lua_State *L);

int PGExecute(lua_State *L)
{
  int tnum,i;
  PQTags *t = (PQTags*)(lua_touserdata(L,-1));
  /* first param is a PGconn table, second is the SQL to run */
  PGconn *p = 0;
  PGresult *r;
  lua_pushstring(L,"PGconn");
  lua_gettable(L,1);
  if (t->connectiontag != lua_tag(L,-1))
    {
      lua_error(L,"Unable to execute a postgres SQL command on a non-postgres connection");
    }
  p = (PGconn*)(lua_touserdata(L,-1));
  if (!lua_isstring(L,2))
    {
      lua_error(L,"Unable to execute non-string queries");
    }
  r = PQexec(p,lua_tostring(L,2));
  /* We return... nil,msg on error */
  /* "cmd",nil for command okay */
  /* "res",table for tuple returns */
  switch(PQresultStatus(r))
    {
    case PGRES_EMPTY_QUERY:
    case PGRES_BAD_RESPONSE:
    case PGRES_NONFATAL_ERROR:
    case PGRES_FATAL_ERROR:
      lua_pushnil(L);
      lua_pushstring(L,PQresStatus(PQresultStatus(r)));
      PQclear(r);
      return 2; /* top 2 elements */
      break;
    case PGRES_COMMAND_OK:
      lua_pushstring(L,"cmd");
      lua_pushnil(L);
      PQclear(r);
      return 2;
      break;
    case PGRES_COPY_OUT:
    case PGRES_COPY_IN:
      lua_pushnil(L);
      lua_pushstring(L,"ENOTIMPLEMENTED");
      PQclear(r);
      return 2;
      break;
    case PGRES_TUPLES_OK:
      /* do nothing for now */
      break;
    default:
      lua_pushnil(L);
      lua_pushstring(L,"Unknown return code.");
      PQclear(r);
      return 2;
      break;
    }
  /* Okay then, we have tuples, fill out the damned structure */
  lua_settop(L,0); /* Clear the stack */

  lua_pushstring(L,"res"); /* return string */
  lua_newtable(L); /* Result table  - Offset 2 on stack */
  
  lua_pushstring(L,"PGresult");
  lua_pushusertag(L, r, t->resulttag);
  lua_settable(L,2);

  lua_pushstring(L,"Done");
  lua_pushusertag(L, t, t->structtag);
  lua_pushcclosure(L, PGDone, 1);
  lua_settable(L,2);
  
  lua_pushstring(L, "rows");
  lua_pushnumber(L, PQntuples(r));
  lua_settable(L,2);

  lua_pushstring(L, "cols");
  lua_pushnumber(L, PQnfields(r));
  lua_settable(L,2);

  lua_pushstring(L,"fields");
  lua_newtable(L); /* Field names */
  tnum = lua_gettop(L); /* for settable() calls */
  lua_pushstring(L,"byname");
  lua_newtable(L); /* byname */
  lua_pushstring(L,"byindex");
  lua_newtable(L); /* byindex */


  for(i=0; i < PQnfields(r);i++)
    {
      lua_pushstring(L,PQfname(r,i));
      lua_pushnumber(L,i);
      lua_settable(L,tnum+2);
      lua_pushnumber(L,i);
      lua_pushstring(L,PQfname(r,i));
      lua_settable(L,tnum+4);
    }

  lua_pushstring(L,"n"); /* Relies on there not being a field called 'n' */
  lua_pushnumber(L,PQnfields(r));
  lua_settable(L,tnum+4);
  
  lua_settable(L,tnum);
  lua_settable(L,tnum);

  lua_settable(L,2); /* put that fields table into res.fields */

  lua_pushstring(L,"get");
  lua_pushusertag(L, t, t->structtag);
  lua_pushcclosure(L, PGGetRow, 1);
  lua_settable(L,2);

  /* stack contains nil,res */
  return 2;
}

int PGGetRow(lua_State *L)
{
  PQTags *t = (PQTags*)(lua_touserdata(L,-1));
  PGresult *r;
  lua_pushstring(L,"PGresult");
  lua_gettable(L,1);
  
  if (t->resulttag != lua_tag(L,-1))
    {
      lua_error(L,"Unable to fetch results from non-postgres results");
    }
  
  /* sanity check params */
  if (!lua_isnumber(L,2) || !lua_isnumber(L,3))
    {
      lua_error(L,"Expecting two numeric params");
    }
  r = (PGresult*)(lua_touserdata(L,-1));
  
  if (lua_tonumber(L,2) < 0 || lua_tonumber(L,2) >= PQntuples(r))
    {
      lua_error(L,"Row request out of range");
    }
  if (lua_tonumber(L,3) < 0 || lua_tonumber(L,3) >= PQnfields(r))
    {
      lua_error(L,"Field request out of range");
    }
  lua_pushstring(L,PQgetvalue(r,lua_tonumber(L,2),lua_tonumber(L,3)));
  return 1; /* the string */
}


